# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Logging
import logging
# Operative system
# import signal
import re
import os
import sys
import stat
import shlex
import subprocess as sp
from copy import deepcopy
from grp import getgrnam
from shutil import copyfile, rmtree
from multiprocessing import Process, Queue, Event, Value
from multiprocessing.managers import SyncManager

# jetson_stats imports
from .core.exceptions import JtopException
from .core.common import get_key, get_var, get_uptime
from .core.hardware import get_hardware, get_platform_variables
from .core.command import Command
from .core.config import Config
from .core.timer_reader import TimerReader
from .core.cpu import CPUService
from .core.memory import MemoryService
from .core.processes import ProcessService
from .core.gpu import GPUService
from .core.engine import EngineService
from .core.temperature import TemperatureService
from .core.power import PowerService
from .core.fan import FanService
from .core.jetson_clocks import JetsonClocksService
from .core.nvpmodel import NVPModelService
# Create logger
logger = logging.getLogger(__name__)
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError
# Load queue library for python 2 and python 3
try:
    import queue
except ImportError:
    import Queue as queue  # pyright: ignore[reportMissingImports]
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)

# Pipe configuration
# https://refspecs.linuxfoundation.org/FHS_3.0/fhs/ch05s13.html
# https://en.wikipedia.org/wiki/Filesystem_Hierarchy_Standard
JTOP_PIPE = '/run/jtop.sock'
JTOP_USER = 'jtop'
JTOP_SERVICE_NAME = 'jtop.service'
# Gain timeout lost connection
TIMEOUT_GAIN = 3
TIMEOUT_SWITCHOFF = 3.0


def status_service(service=JTOP_SERVICE_NAME):
    return os.system('systemctl is-active --quiet {service}'.format(service=service)) == 0


def remove_service_pipe():
    # Remove old pipes if exists
    if os.path.isdir(JTOP_PIPE):
        logger.info("Remove folder {pipe}".format(pipe=JTOP_PIPE))
        rmtree(JTOP_PIPE)
    elif os.path.exists(JTOP_PIPE):
        logger.info("Remove pipe {pipe}".format(pipe=JTOP_PIPE))
        os.remove(JTOP_PIPE)


def uninstall_service(name=JTOP_SERVICE_NAME):
    if os.path.isfile('/etc/systemd/system/{name}'.format(name=name)) or os.path.islink('/etc/systemd/system/{name}'.format(name=name)):
        logger.info("Found {name}".format(name=name))
        # Check if service is active
        if status_service(service=name):
            # Stop service
            logger.info(" - STOP {name}".format(name=name))
            sp.call(shlex.split('systemctl stop {name}'.format(name=name)))
        # Disable service
        logger.info(" - DISABLE {name}".format(name=name))
        sp.call(shlex.split('systemctl disable {name}'.format(name=name)))
        # Remove jetson_performance service from /etc/init.d
        if os.path.isfile('/etc/systemd/system/{name}'.format(name=name)) or os.path.islink('/etc/systemd/system/{name}'.format(name=name)):
            logger.info(" - REMOVE {name} from /etc/systemd/system".format(name=name))
            os.remove('/etc/systemd/system/{name}'.format(name=name))
        # Update service list
        logger.info(" - Reload all daemons")
        sp.call(shlex.split('systemctl daemon-reload'))
        return True
    return False


def install_service(package_root, copy, name=JTOP_SERVICE_NAME):
    logger.info("Install {name}".format(name=name))
    # Copy or link file
    service_install_path = '/etc/systemd/system/{name}'.format(name=name)
    service_package_path = '{package_root}/services/{name}'.format(package_root=package_root, name=name)
    # remove if exist file
    if os.path.isfile(service_install_path) or os.path.islink(service_install_path):
        logger.info(" - Remove old {path}".format(path=service_install_path))
        os.remove(service_install_path)
    if copy:
        type_service = "Copying"
        copyfile(service_package_path, service_install_path)
    else:
        type_service = "Linking"
        os.symlink(service_package_path, service_install_path)
    # Prompt message
    logger.info(" - {type} {file} -> {path}".format(type=type_service.upper(), file=name, path=service_install_path))
    # Update service list
    cmd_daemon_reload = Command(shlex.split('systemctl daemon-reload'))
    try:
        cmd_daemon_reload()
        logger.info(" - Reload all daemons")
    except (OSError, Command.CommandException):
        logger.error("Fail reload all daemons")
    # Enable jetson_stats at startup
    cmd_service_enable = Command(shlex.split('systemctl enable {name}'.format(name=name)))
    try:
        cmd_service_enable()
        logger.info(" - ENABLE {name}".format(name=name))
        # logger.info(lines)
    except (OSError, Command.CommandException):
        logger.error("Fail enable service {name}".format(name=name))
    # Start service
    cmd_service_start = Command(shlex.split('systemctl start {name}'.format(name=name)))
    try:
        cmd_service_start()
        logger.info(" - START {name}".format(name=name))
        # logger.info(lines)
    except (OSError, Command.CommandException):
        logger.error("Fail start service {name}".format(name=name))


def status_permission_user(group=JTOP_USER):
    user = os.environ.get('USER', '')
    # Get user from sudo
    sudo_user = os.environ.get('SUDO_USER', '')
    # If are both empty assign 'root'
    user = sudo_user or 'root'
    # Check if user is in group
    cmd_group_user = Command(shlex.split('groups {user}'.format(user=user)))
    try:
        lines = cmd_group_user()
        for line in lines:
            if ":" not in line:
                continue
            name, info = line.split(":")
            info = info.strip().split()
            if name.strip() == user and group in info:
                return True
    except (OSError, Command.CommandException):
        logger.error("{user} does not exist".format(user=user))
    return False


def status_permission_group(group=JTOP_USER):
    # Check if exist group
    cmd_group = Command(shlex.split('getent group {group}'.format(group=group)))
    try:
        cmd_group()
    except (OSError, Command.CommandException):
        logger.error("Does not exist {group}".format(group=group))
        return False
    return True


def status_permission(group=JTOP_USER):
    return status_permission_user(group) and status_permission_group(group)


def unset_service_permission(group=JTOP_USER):
    user = os.environ.get('USER', '')
    # Get user from sudo
    sudo_user = os.environ.get('SUDO_USER', '')
    # If are both empty assign 'root'
    user = sudo_user or 'root'
    # Check if user is in group
    if status_permission_user(group):
        logger.info("Remove {user} from group {group}".format(group=group, user=user))
        sp.call(shlex.split('gpasswd -d {user} {group}'.format(group=group, user=user)))
    if status_permission_group(group):
        logger.info("Delete group {group}".format(group=group))
        sp.call(shlex.split('groupdel {group}'.format(group=group)))


def set_service_permission(group=JTOP_USER):
    user = os.environ.get('USER', '')
    # Get user from sudo
    sudo_user = os.environ.get('SUDO_USER', '')
    # If are both empty assign 'root'
    user = sudo_user or 'root'
    # Make jetson_stats group
    if not status_permission_group(group):
        logger.info("Add new group {group}".format(group=group))
        sp.call(shlex.split('groupadd {group}'.format(group=group)))
    if not status_permission_user(group):
        logger.info("Add {user} to group {group}".format(group=group, user=user))
        sp.call(shlex.split('usermod -a -G {group} {user}'.format(group=group, user=user)))


class JtopManager(SyncManager):

    def __init__(self, authkey=None):
        if authkey is None:
            authkey = get_key().encode("utf-8")
        super(JtopManager, self).__init__(address=(JTOP_PIPE), authkey=authkey)

    def get_queue(self):
        pass

    def sync_data(self):
        pass

    def sync_event(self):
        pass


class JtopServer(Process):
    """
        - https://pymotw.com/2/multiprocessing/basics.html
        - https://pymotw.com/2/multiprocessing/communication.html
        - https://stackoverflow.com/questions/1829116/how-to-share-variables-across-scripts-in-python
        - https://stackoverflow.com/questions/45342200/how-to-use-syncmanager-lock-or-event-correctly
        - https://stackoverflow.com/questions/2545961/how-to-synchronize-a-python-dict-with-multiprocessing
        - https://docs.python.org/2.7/reference/datamodel.html
    """

    def __init__(self, force=False):
        self.force = force
        # Check if running a root
        if os.getuid() != 0:
            raise JtopException("jtop service need sudo to work")
        # Load configuration
        self.config = Config()
        # Save version jtop
        self._version = deepcopy(get_var(VERSION_RE))
        logger.info("jetson_stats {version} - server loaded".format(version=self._version))
        # Error queue
        self._error = Queue()
        # Command queue
        self.q = Queue()
        # Speed interval
        self.interval = Value('d', -1.0)
        # Dictionary to sync
        self.data = {}
        # Event lock
        self.event = Event()
        # Load super Thread constructor
        super(JtopServer, self).__init__()
        # Register stats
        # https://docs.python.org/2/library/multiprocessing.html#using-a-remote-manager
        JtopManager.register('get_queue', callable=lambda: self.q)
        JtopManager.register("sync_data", callable=lambda: self.data)
        JtopManager.register('sync_event', callable=lambda: self.event)
        # Generate key and open broadcaster
        self.broadcaster = JtopManager()
        # Load board and platform variables
        data_platform = get_platform_variables()
        logger.info("Running on Python: {python_version}".format(python_version=data_platform['Python']))
        self.board = {'hardware': get_hardware()}
        # From this point are initialized or hardware services
        # Setup cpu service
        self.cpu = CPUService()
        # Setup gpu service
        self.gpu = GPUService()
        # Setup process service
        self.processes = ProcessService()
        # Setup memory service
        self.memory = MemoryService(self.config)
        # Setup engine service
        self.engine = EngineService()
        # Setup Temperature service
        self.temperature = TemperatureService()
        # Setup Power meter service
        self.power = PowerService()
        # Initialize Fan
        self.fan = FanService(self.config)
        # Initialize jetson_clocks controller
        self.jetson_clocks = JetsonClocksService(self.config, self.fan)
        # Initialize nvpmodel controller
        self.nvpmodel = NVPModelService(self.jetson_clocks)
        # Initialize timer reader
        self._timer_reader = TimerReader(self.jtop_stats)

    def run(self):
        logger.info("Initialization service")
        # Initialize jetson_fan
        self.fan.initialization()
        # Run setup
        if self.jetson_clocks.exists():
            # Decode for initialization and reset
            data = self.jtop_decode()
            # Reset CPU estimation
            self.cpu.reset_estimation()
            # Reset avg temperatures
            self.power.reset_avg_power()
            # Initialization jetson_clocks
            self.jetson_clocks.initialization(self.nvpmodel, data)
        logger.info("service ready")
        # Initialize variables
        timeout = None
        interval = 1
        try:
            while True:
                try:
                    # Decode control message
                    control = self.q.get(timeout=timeout)
                    # Check if control is not empty
                    if not control:
                        continue
                    # If service read the same init message resend it
                    if 'init' in control:
                        self.q.put(control)
                        continue
                    logger.debug("control message {control}".format(control=control))
                    # Manage swap
                    if 'swap' in control:
                        swap = control['swap']
                        if 'command' in swap:
                            command = swap['command']
                            if command == 'set':
                                logger.info("Activating swap in {path}".format(path=swap['path']))
                                self.memory.swap_set(swap['size'], swap['path'], swap['boot'])
                            elif command == 'unset':
                                logger.info("Deactivating swap in {path}".format(path=swap['path']))
                                self.memory.swap_deactivate(swap['path'])
                            else:
                                logger.error("swap command not detected: {command}".format(command=command))
                        else:
                            logger.error("no swap command in this message {message}".format(message=swap))
                    # Clear cache
                    if 'clear_cache' in control:
                        # Clear cache
                        self.memory.clear_cache()
                        logger.info("Clear cache")
                    # Set GPU configuration
                    if 'gpu' in control:
                        gpu = control['gpu']
                        if 'command' in gpu:
                            command = gpu['command']
                            name = gpu['name']
                            if command == '3d_scaling':
                                self.gpu.set_scaling_3D(name, gpu['value'])
                            elif command == 'railgate':
                                self.gpu.set_railgate(name, gpu['value'])
                            else:
                                logger.error("gpu command not detected: {command}".format(command=command))
                        else:
                            logger.error("no gpu command in this message {message}".format(message=gpu))
                    # Speed Fan and configuration
                    if 'fan' in control:
                        fan = control['fan']
                        if 'command' in fan:
                            command = fan['command']
                            if command == 'profile':
                                name = fan['name']
                                profile = fan['profile']
                                logger.info('Fan \"{name}\" set profile {profile}'.format(name=name, profile=profile))
                                self.fan.set_profile(name, profile)
                            elif command == 'speed':
                                name = fan['name']
                                speed = fan['speed']
                                idx = fan['idx']
                                logger.info('Fan \"{name}[{idx}]\" set speed {speed}'.format(name=name, idx=idx, speed=speed))
                                self.fan.set_speed(name, speed, idx)
                            else:
                                logger.error("fan command not detected: {command}".format(command=command))
                        else:
                            logger.error("no fan command in this message {message}".format(message=fan))
                    if 'jc' in control:
                        jc = control['jc']
                        # Enable / disable jetson_clocks
                        if 'enable' in jc:
                            self.jetson_clocks.set_enable(jc['enable'])
                        # Update jetson_clocks configuration
                        if 'boot' in jc:
                            self.jetson_clocks.set_boot(jc['boot'])
                        # Clear configuration
                        if 'clear' in jc:
                            self.jetson_clocks.clear()
                    # Decode nvp model
                    if 'nvp' in control:
                        nvpmodel = control['nvp']
                        nvpmodel_id = nvpmodel['id']
                        nvpmodel_force = nvpmodel['force']
                        # Set new NV Power Mode
                        logger.info("Set new NV Power Model ID {id}".format(id=nvpmodel_id))
                        self.nvpmodel.set_nvpmodel_id(nvpmodel_id, nvpmodel_force)
                    # Initialize tegrastats speed
                    if 'interval' in control:
                        interval = control['interval']
                        # Run stats
                        if self._timer_reader.open(interval=interval):
                            # Set interval value
                            self.interval.value = interval
                            # Status start tegrastats
                            logger.info("jtop timer thread started {interval}ms".format(interval=int(interval * 1000)))
                        # send configuration board
                        init = {
                            'version': self._version,
                            'board': self.board,
                            'interval': self.interval.value,
                            'memory': self.memory.swap_path(),
                            'fan': self.fan.get_configs(),
                            'jc': self.jetson_clocks.exists(),
                        }
                        # If nvpmodel exist load all modes
                        if self.nvpmodel.exists():
                            init['nvpmodel'] = {
                                'models': self.nvpmodel.get_all_nvpmodels(),
                                'default': self.nvpmodel.get_default()
                            }
                        # Push initialization string
                        self.q.put({'init': init})
                    # Manage jetson_clocks
                    if 'config' in control:
                        command = control['config']
                        if command == 'reset':
                            logger.info('Reset configuration')
                            self.config.clear()
                    # Update timeout interval
                    timeout = TIMEOUT_GAIN if interval <= TIMEOUT_GAIN else interval * TIMEOUT_GAIN
                except queue.Empty:
                    self.sync_event.clear()
                    # Reset CPU estimation
                    self.cpu.reset_estimation()
                    # Reset avg temperatures
                    self.power.reset_avg_power()
                    # Close and log status
                    if self._timer_reader.close():
                        logger.info("jtop timer thread close")
                    # Disable timeout
                    timeout = None
                    self.interval.value = -1.0
        except (KeyboardInterrupt, SystemExit):
            logger.warning("KeyboardInterrupt, SystemExit interrupt")
        except FileNotFoundError:
            logger.warning("FileNotFoundError on service thread")
        except Exception as e:
            logger.error("Error subprocess {error}".format(error=e), exc_info=1)
            # Write error message
            self._error.put(sys.exc_info())
        finally:
            # Close tegra
            if self._timer_reader.close(timeout=TIMEOUT_SWITCHOFF):
                logger.info("FORCE jtop timer thread close")
                # Reset CPU estimation
                self.cpu.reset_estimation()
                # Reset avg temperatures
                self.power.reset_avg_power()

    def start(self):
        # Initialize socket
        try:
            gid = getgrnam(JTOP_USER).gr_gid
        except KeyError:
            # User does not exist
            raise JtopException("Group {jtop_user} does not exist!".format(jtop_user=JTOP_USER))
        # Remove old pipes if exists
        if self.force:
            self.remove_files()
        # If exist JTOP pipe raise exception
        if os.path.exists(JTOP_PIPE):
            raise JtopException("Service already active! Please check before run it again")
        # Start broadcaster
        try:
            self.broadcaster.start()
        except EOFError:
            raise JtopException("Server already alive")
        # Initialize synchronized data and conditional
        self.sync_data = self.broadcaster.sync_data()
        self.sync_event = self.broadcaster.sync_event()
        # Change owner
        os.chown(JTOP_PIPE, os.getuid(), gid)
        # Change mode controller and stats
        # https://www.tutorialspoint.com/python/os_chmod.htm
        # Equivalent permission 660 srw-rw----
        os.chmod(JTOP_PIPE, stat.S_IREAD | stat.S_IWRITE | stat.S_IWGRP | stat.S_IRGRP)
        # Run the Control server
        super(JtopServer, self).start()

    def loop_for_ever(self):
        try:
            self.start()
        except JtopException as e:
            logger.error(e)
            return
        # Join main subprocess
        try:
            self.join()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            # Close communication
            self.close()

    def close(self):
        self.q.close()
        self.broadcaster.shutdown()
        # If process is alive wait to quit
        # logger.debug("Status subprocess {status}".format(status=self.is_alive()))
        while self.is_alive():
            # If process is in timeout manually terminate
            if self.interval.value == -1.0:
                logger.info("Terminate subprocess")
                self.terminate()
            logger.info("Wait shutdown subprocess")
            self.join(timeout=TIMEOUT_SWITCHOFF)
            self.interval.value = -1.0
        # Close tegrastats
        try:
            error = self._error.get(timeout=0.5)
            # Raise error if exist
            if error:
                ex_type, ex_value, tb_str = error
                ex_value.__traceback__ = tb_str
                raise ex_value
        except queue.Empty:
            pass
        self.remove_files()
        # Switch off jetson_clocks if there are threads alive
        self.jetson_clocks.close()
        # Switch off nvpmodel if there are threads alive
        self.nvpmodel.close()
        # Close stats server
        logger.info("Service closed")
        return True

    def remove_files(self):
        # If exist remove pipe
        remove_service_pipe()

    def config_clear(self):
        self.config.clear()

    def jtop_decode(self):
        # Make configuration dict
        data = {}
        # -- UPTIME --
        data['uptime'] = get_uptime()
        # -- CPU --
        # Read CPU data
        data['cpu'] = self.cpu.get_status()
        # -- GPU ---
        # Read GPU status
        data['gpu'] = self.gpu.get_status()
        # -- All processes
        total, table = self.processes.get_status()
        data['processes'] = table
        # -- RAM --
        # Read memory data
        # In this dictionary are collected
        # - RAM
        # - SWAP
        # - EMC (If available)
        # - IRAM (If available)
        data['mem'] = self.memory.get_status(total)
        # -- Engines --
        # Read all engines available
        # Can be empty for x86 architecture
        data['engines'] = self.engine.get_status()
        # -- Temperature --
        data['temperature'] = self.temperature.get_status()
        # -- Power --
        data['power'] = self.power.get_status()
        # -- FAN --
        # Update status fan speed
        data['fan'] = self.fan.get_status()
        # -- JETSON_CLOCKS --
        # Jetson Clock to check if is running need to read the status from:
        # CPU, EMC, GPU, ENGINES
        # If not empty return status
        if self.jetson_clocks.exists():
            # Get status jetson_clocks
            data['jc'] = self.jetson_clocks.get_status(data)
        # -- NVP MODEL --
        # Read nvpmodel status, ID, name, and status thread
        if self.nvpmodel.exists():
            data['nvp'] = self.nvpmodel.get_status()
        return data

    def jtop_stats(self):
        # logger.info("jtop read")
        data = self.jtop_decode()
        # Pack and send all data
        # https://stackoverflow.com/questions/6416131/add-a-new-item-to-a-dictionary-in-python
        self.sync_data.update(data)
        # Set event for all clients
        if not self.sync_event.is_set():
            self.sync_event.set()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from os import path
import curses
# Page class definition
from .jtopgui import Page
# Graphics elements
from .lib.colors import NColors
from .lib.common import (size_min, unit_to_string, size_to_string, plot_name_info)
from .lib.linear_gauge import basic_gauge, basic_gauge_simple
from .lib.chart import Chart
from .lib.smallbutton import SmallButton
from .pcpu import cpu_grid

SWAP_MAX_SIZE = 100
SWAP_MIN_SIZE = 1


def mem_gauge(stdscr, pos_y, pos_x, size, mem_data):
    # Plot values
    values = [
        ((mem_data['used'] - mem_data['shared']) / mem_data['tot'] * 100.0, NColors.cyan()),
        (mem_data['shared'] / mem_data['tot'] * 100.0, NColors.green() | curses.A_BOLD),
        (mem_data['buffers'] / mem_data['tot'] * 100.0, NColors.blue()),
        (mem_data['cached'] / mem_data['tot'] * 100.0, NColors.yellow()),
    ]
    used = size_to_string(mem_data['used'], 'k')
    total = size_to_string(mem_data['tot'], 'k')
    # Draw gauge
    data = {
        'name': 'Mem',
        'color': NColors.cyan(),
        'values': values,
        'mright': "{used}/{total}".format(used=used, total=total)
    }
    # Draw gauge
    basic_gauge(stdscr, pos_y, pos_x, size - 1, data)


def swap_gauge(stdscr, pos_y, pos_x, size, mem_data):
    # Plot values
    values = [
        (mem_data['used'] / mem_data['tot'] * 100.0, NColors.red()),
        (mem_data['cached'] / mem_data['tot'] * 100.0, NColors.yellow()),
    ] if mem_data['tot'] > 0 else []
    used = size_to_string(mem_data['used'], 'k')
    total = size_to_string(mem_data['tot'], 'k')
    # Draw gauge
    data = {
        'name': 'Swp',
        'color': NColors.cyan(),
        'values': values,
        'online': mem_data['tot'] > 0,
        'mright': "{used}/{total}".format(used=used, total=total)
    }
    # Draw gauge
    basic_gauge(stdscr, pos_y, pos_x, size - 1, data)


def emc_gauge(stdscr, pos_y, pos_x, size, mem_data):
    # online status
    online = mem_data['online'] if 'online' in mem_data else True
    # Plot values
    if 'max' in mem_data:
        values = [
            (mem_data['cur'] / mem_data['max'] * 100.0, NColors.green()),
        ] if online else []
        # Draw gauge
        data = {
            'name': 'Emc',
            'color': NColors.cyan(),
            'values': values,
            'mleft': unit_to_string(mem_data['min'], 'k', 'Hz') if 'min' in mem_data else '',
            'mright': unit_to_string(mem_data['max'], 'k', 'Hz'),
        }
        # Draw gauge
        basic_gauge(stdscr, pos_y, pos_x, size - 13, data, bar=':')
        # Draw info EMC
        curr_string = unit_to_string(mem_data['cur'], 'k', 'Hz')
        stdscr.addstr(pos_y, pos_x + size - 11, curr_string, NColors.italic())
    else:
        mem_data['name'] = 'Emc'
        basic_gauge_simple(stdscr, pos_y, pos_x, size - 6, mem_data)
    # Show val
    curr_string = "{val:3.0f}%".format(val=mem_data['val'])
    stdscr.addstr(pos_y, pos_x + size - 4, curr_string, curses.A_BOLD)


def iram_gauge(stdscr, pos_y, pos_x, size, mem_data):
    used = size_to_string(mem_data['used'], 'k')
    total = size_to_string(mem_data['tot'], 'k')
    # Make data for gauge
    data = {
        'name': 'Iram',
        'color': NColors.cyan(),
        'values': [(mem_data['used'] / mem_data['tot'] * 100.0, NColors.cyan())],
        'mright': "{used}/{total}".format(used=used, total=total),
    }
    basic_gauge(stdscr, pos_y, pos_x, size - 12, data)
    # Write lfb
    label_lfb = "(lfb {nblock}{unit}B)".format(nblock=mem_data['lfb'], unit='k')
    stdscr.addstr(pos_y, pos_x + size - 11, label_lfb, curses.A_NORMAL)


def compact_memory(stdscr, pos_y, pos_x, width, height, jetson):
    line_counter = 1
    # Draw memory gauge
    try:
        mem_gauge(stdscr, pos_y, pos_x, width, jetson.memory['RAM'])
    except curses.error:
        return 0
    try:
        swap_gauge(stdscr, pos_y + line_counter, pos_x, width, jetson.memory['SWAP'])
        line_counter += 1
    except curses.error:
        return line_counter
    if 'EMC' in jetson.memory:
        try:
            emc_gauge(stdscr, pos_y + line_counter, pos_x, width, jetson.memory['EMC'])
            line_counter += 1
        except curses.error:
            return line_counter
    if 'IRAM' in jetson.memory:
        try:
            iram_gauge(stdscr, pos_y + line_counter, pos_x, width, jetson.memory['IRAM'])
            line_counter += 1
        except curses.error:
            pass
    return line_counter


class MEM(Page):

    def __init__(self, stdscr, jetson):
        super(MEM, self).__init__("MEM", stdscr, jetson)
        # Initialize MEM chart
        self.chart_ram = Chart(jetson, "RAM", self.update_chart,
                               type_value=float,
                               color_text=curses.COLOR_CYAN,
                               color_chart=[curses.COLOR_YELLOW, curses.COLOR_BLUE, curses.COLOR_GREEN, curses.COLOR_CYAN])
        # Initialize buttons
        self._button_cache = SmallButton(stdscr, self.action_cache, "clear cache", trigger_key='c')
        self._button_increase = SmallButton(stdscr, self.action_increase, trigger_key='+')
        self._button_decrease = SmallButton(stdscr, self.action_decrease, trigger_key='-')
        self._button_create = SmallButton(stdscr, self.action_create, "Create new", trigger_key='s')
        self._button_boot = SmallButton(stdscr, self.action_boot, "on boot", toggle=True, trigger_key='b')
        self._button_swap = SmallButton(stdscr, self.action_swap)
        # Size swap
        self._swap_size = SWAP_MIN_SIZE
        self._swap_old_size = self._swap_size
        self._on_boot = False
        # Swap table control
        self._swap_pressed = -1
        self._swap_name = self.get_new_swap_name()

    def action_cache(self, info, selected):
        self.jetson.memory.clear_cache()

    def action_create(self, info, selected):
        # Change status swap
        if not self.jetson.memory.swap_is_enable(self._swap_name):
            self.jetson.memory.swap_set(self._swap_size, self._swap_name, on_boot=self._on_boot)
            self._swap_old_size = self._swap_size

    def action_boot(self, info, selected):
        self._on_boot = selected

    def action_swap(self, info, selected):
        if self._swap_pressed != -1:
            # Read name swap
            name = list(self._swaps)[self._swap_pressed]
            swap = self._swaps[name]
            # Deactivate SWAP
            if self.jetson.memory.swap_is_enable(name) and swap['type'] == 'file':
                self.jetson.memory.swap_deactivate(name)

    def action_increase(self, info, selected):
        if self._swap_size < SWAP_MAX_SIZE:
            self._swap_size += 1

    def action_decrease(self, info, selected):
        if self._swap_size > SWAP_MIN_SIZE:
            self._swap_size -= 1

    def update_chart(self, jetson, name):
        parameter = jetson.memory['RAM']
        # Get max value if is present
        max_val = parameter.get("tot", 100)
        # Get value
        cached_val = parameter.get("cached", 0)
        buffer_val = parameter.get("buffers", 0)
        use_val = parameter.get("used", 0)
        cpu_val = parameter.get("used", 0) - parameter.get("shared", 0)
        szw, divider, unit = size_min(max_val, start='k')
        # Append in list
        cached_out = (cached_val + buffer_val + use_val) / divider
        buffers_out = (buffer_val + use_val) / divider
        gpu_out = (use_val) / divider
        cpu_out = cpu_val / divider
        return {
            'value': [cached_out, buffers_out, gpu_out, cpu_out],
            'max': szw,
            'unit': unit
        }

    def draw_ram_legend(self, pos_y, pos_x):
        self.stdscr.addstr(pos_y, pos_x + 1, "     RAM     ", curses.A_REVERSE)
        # Plot all RAM values
        used = size_to_string(self.jetson.memory['RAM']['used'], 'k')
        plot_name_info(self.stdscr, pos_y + 1, pos_x + 1, 'Used', used, spacing=3, color=NColors.cyan())
        shared = size_to_string(self.jetson.memory['RAM']['shared'], 'k')
        plot_name_info(self.stdscr, pos_y + 2, pos_x + 1, 'GPU Sh', shared, spacing=1, color=NColors.green())
        buffers = size_to_string(self.jetson.memory['RAM']['buffers'], 'k')
        plot_name_info(self.stdscr, pos_y + 3, pos_x + 1, 'Buffers', buffers, color=NColors.blue())
        cached = size_to_string(self.jetson.memory['RAM']['cached'], 'k')
        plot_name_info(self.stdscr, pos_y + 4, pos_x + 1, 'Cached', cached, spacing=1, color=NColors.yellow())
        free = size_to_string(self.jetson.memory['RAM']['free'], 'k')
        plot_name_info(self.stdscr, pos_y + 5, pos_x + 1, 'Free', free, spacing=3)
        total = size_to_string(self.jetson.memory['RAM']['tot'], 'k')
        plot_name_info(self.stdscr, pos_y + 6, pos_x + 1, 'TOT', total, spacing=4, color=curses.A_BOLD)

    def print_zram(self, stdscr, idx, swap, pos_y, pos_x, size_h, size_w):
        name, swap = swap
        # data gauge
        used = size_to_string(swap['used'], 'k')
        total = size_to_string(swap['size'], 'k')
        data = {
            'name': path.basename(name),
            'color': NColors.cyan(),
            'values': [(swap.get('used', 0) / swap.get('size', 0) * 100.0, NColors.cyan())],
            'mleft': "P{prio}".format(prio=swap['prio']),
            'mright': "{used}/{total}".format(used=used, total=total),
        }
        try:
            basic_gauge(stdscr, pos_y, pos_x, size_w - 2, data)
        except curses.error:
            pass

    def draw_swap_table(self, pos_y, pos_x, width, height, key, mouse):
        swap_info = self.jetson.memory['SWAP']
        used = size_to_string(swap_info['used'], 'k')
        total = size_to_string(swap_info['tot'], 'k')
        cached = size_to_string(swap_info['cached'], 'k')
        self.stdscr.addstr(pos_y, pos_x + 1, "SWAP", curses.A_BOLD)
        self.stdscr.addstr(pos_y, pos_x + 6, "{used}/{total} (Cached {cached})".format(used=used, total=total, cached=cached), NColors.red())
        # Print only zrams
        _, _, _, size_rows = cpu_grid(self.stdscr, self._zrams.items(), self.print_zram, pos_y + 1, pos_x + 1, size_width=width)
        # swap position
        pos_y_swap = pos_y + size_rows + 1
        # Detect line pressed
        line_pressed = self._mousePressed(mouse, pos_y_swap, pos_x, width, len(self._swaps) - 1)
        if line_pressed is not None:
            self._swap_pressed = line_pressed
        # Draw swap table
        for idx, (name, swap) in enumerate(self._swaps.items()):
            if pos_y_swap + idx < height:
                # Color gauge
                # Change color for type partition
                if swap['type'] == 'partition':
                    color = NColors.magenta() if self._swap_pressed != idx else NColors.imagenta()
                elif swap['type'] == 'file':
                    color = NColors.yellow() if self._swap_pressed != idx else NColors.iyellow()
                else:
                    color = NColors.green() if self._swap_pressed != idx else NColors.igreen()
                # data gauge
                used = size_to_string(swap['used'], 'k')
                total = size_to_string(swap['size'], 'k')
                data = {
                    'name': path.basename(name),
                    'color': color,
                    'values': [(swap.get('used', 0) / swap.get('size', 0) * 100.0, color)],
                    'mleft': "P{prio}".format(prio=swap['prio']),
                    'mright': "{used}/{total}".format(used=used, total=total),
                }
                width_swap = width - 8 if swap['boot'] else width - 3
                basic_gauge(self.stdscr, pos_y_swap + idx, pos_x + 1, width_swap, data)
                if swap['boot']:
                    color_boot = curses.A_BOLD if self._swap_pressed != idx else curses.A_REVERSE | curses.A_BOLD
                    self.stdscr.addstr(pos_y_swap + idx, pos_x + width_swap + 3, "Boot", color_boot)
            else:
                for n_arrow in range(width):
                    self.stdscr.addch(height, pos_x + n_arrow, curses.ACS_DARROW, curses.A_REVERSE | curses.A_BOLD)
                break

    def draw_swap_controller(self, pos_y, pos_x, key, mouse):
        string_name = ''
        color = curses.A_REVERSE
        # Read swap name
        if self._swap_pressed != -1:
            operation = ''
            # Read name swap
            name = list(self._swaps)[self._swap_pressed]
            # Get info swap
            swap = self._swaps[name]
            if swap['type'] == 'partition':
                color = NColors.imagenta()
            elif swap['type'] == 'file':
                color = NColors.iyellow()
                operation = 'Disable '
            else:
                color = NColors.icyan()
            # Swap name and operation
            string_name = "{operation}{name}".format(operation=operation, name=name)
        # Swap controller button
        try:
            label = string_name if string_name else 'Select swap'
            self._button_swap.update(pos_y, pos_x, label, key, mouse, color=color)
        except curses.error:
            pass
        # Button create new swap
        try:
            self._button_create.update(pos_y + 2, pos_x, key=key, mouse=mouse)
            self._button_boot.update(pos_y + 3, pos_x, key=key, mouse=mouse)
        except curses.error:
            pass
        # Draw selected number
        try:
            if self._swap_size > self._swap_old_size:
                color = NColors.green()
            elif self._swap_size < self._swap_old_size:
                color = NColors.yellow()
            else:
                color = curses.A_NORMAL
            self.stdscr.addstr(pos_y + 4, pos_x + 5, "{size: <2}".format(size=self._swap_size), color)
            self.stdscr.addstr(pos_y + 4, pos_x + 8, "GB", curses.A_BOLD)
        except curses.error:
            pass
        # Draw buttons
        try:
            self._button_decrease.update(pos_y + 4, pos_x, key=key, mouse=mouse)
            self._button_increase.update(pos_y + 4, pos_x + 11, key=key, mouse=mouse)
        except curses.error:
            pass
        # Draw swap name
        self._swap_name = self.get_new_swap_name()
        try:
            self.stdscr.addstr(pos_y + 5, pos_x, "New: ")
            self.stdscr.addstr(self._swap_name, NColors.yellow())
        except curses.error:
            pass

    def draw(self, key, mouse):
        swap_info = self.jetson.memory['SWAP']
        swap_table = swap_info['table']
        # Remove all zram
        self._zrams = {}
        self._swaps = {}
        for (name, swap) in swap_table.items():
            if swap['type'] == 'zram':
                self._zrams[name] = swap
            else:
                self._swaps[name] = swap
        # Screen size
        height, width, first = self.size_page()
        # Set size chart memory
        size_x = [1, width - 18]
        size_y = [first + 1, height // 2 - 2]
        # Draw the GPU chart
        # lfb label
        mem_data = self.jetson.memory['RAM']
        used = size_to_string(mem_data['used'], 'k')
        total = size_to_string(mem_data['tot'], 'k')
        percent = "{used}/{total}B".format(used=used, total=total)
        label_lfb = "(lfb {nblock}x4MB)".format(nblock=mem_data['lfb'])
        self.chart_ram.draw(self.stdscr, size_x, size_y, label="{percent} - {lfb}".format(percent=percent, lfb=label_lfb))
        # Draw all RAM
        self.draw_ram_legend(first + 1, width - 16)
        # Draw EMC gague
        line_counter = 0
        if 'EMC' in self.jetson.memory:
            emc_gauge(self.stdscr, first + height // 2 - 1 + line_counter, 1, width - 22, self.jetson.memory['EMC'])
            line_counter += 1
        if 'IRAM' in self.jetson.memory:
            iram_gauge(self.stdscr, first + height // 2 - 1 + line_counter, 1, width - 22, self.jetson.memory['IRAM'])
            line_counter += 1
        # Draw buttons
        self._button_cache.update(first + height // 2, width - 20, key=key, mouse=mouse)
        self.draw_swap_controller(first + height // 2 + 2, width - 20, key, mouse)
        # Draw swap list
        self.draw_swap_table(first + height // 2 - 1 + line_counter, 0, width - 22, height - 2, key, mouse)

    def get_new_swap_name(self):
        # Get default name
        swap_name = self.jetson.memory.swap_path()
        base_name = swap_name
        counter = 0
        # Get table
        swap_info = self.jetson.memory['SWAP']
        swap_table = swap_info['table']
        while swap_name in swap_table:
            swap_name = "{name}_{counter}".format(name=base_name, counter=counter)
            counter += 1
        return swap_name

    def _mousePressed(self, mouse, pos_y, pos_x, width, height):
        line_pressed = None
        if mouse:
            mx, my = mouse
            line_pressed = -1
            if my >= pos_y and my <= pos_y + height and \
               mx >= pos_x and mx <= pos_x + width:
                line_pressed = my - pos_y
        return line_pressed
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
# All regular expressions
DATE_RE = re.compile(r'\d+\-\d+\-\d+ \d+\:\d+\:\d+ ')
SWAP_RE = re.compile(r'SWAP (\d+)\/(\d+)(\w)B( ?)\(cached (\d+)(\w)B\)')
IRAM_RE = re.compile(r'IRAM (\d+)\/(\d+)(\w)B( ?)\(lfb (\d+)(\w)B\)')
RAM_RE = re.compile(r'RAM (\d+)\/(\d+)(\w)B( ?)\(lfb (\d+)x(\d+)(\w)B\)')
MTS_RE = re.compile(r'MTS fg (\d+)% bg (\d+)%')
VALS_RE = re.compile(r'\b([A-Z0-9_]+) ([0-9%@]+)(?=[^/])\b')
VAL_FRE_RE = re.compile(r'\b(\d+)%@(\d+)')
CPU_RE = re.compile(r'CPU \[(.*?)\]')
WATT_RE = re.compile(r'\b(\w+) ([0-9.]+)(\w?)W?\/([0-9.]+)(\w?)W?\b')
TEMP_RE = re.compile(r'\b(\w+)@(-?[0-9.]+)C\b')


def val_freq(val):
    if '@' in val:
        match = VAL_FRE_RE.search(val)
        return {'val': int(match.group(1)), 'frq': int(match.group(2)) * 1000}
    else:
        return {'val': int(val)}


def DATE(text):
    """
        Remove date and time from string
        12-27-2022 13:48:01 ....
    """
    match = DATE_RE.search(text)
    if match is None:
        return text
    else:
        return text[:match.start()] + text[match.end():]


def SWAP(text):
    """
        SWAP X/Y (cached Z)
        X = Amount of SWAP in use in megabytes.
        Y = Total amount of SWAP available for applications.
        Z = Amount of SWAP cached in megabytes.
    """
    match = SWAP_RE.search(text)
    if match:
        return {'use': int(match.group(1)),
                'tot': int(match.group(2)),
                'unit': str(match.group(3)),
                # group 4 is an optional space
                'cached': {'size': int(match.group(5)),
                           'unit': str(match.group(6))}}
    else:
        return {}


def IRAM(text):
    """
        IRAM X/Y (lfb Z)
        IRAM is memory local to the video hardware engine.
        X = Amount of IRAM memory in use, in kilobytes.
        Y = Total amount of IRAM memory available.
        Z = Size of the largest free block.
    """
    match = IRAM_RE.search(text)
    if match:
        return {'use': int(match.group(1)),
                'tot': int(match.group(2)),
                'unit': match.group(3),
                # group 4 is an optional space
                'lfb': {'size': int(match.group(5)),
                        'unit': match.group(6)}}
    else:
        return {}


def RAM(text):
    """
        RAM X/Y (lfb NxZ)
        Largest Free Block (lfb) is a statistic about the memory allocator.
        It refers to the largest contiguous block of physical memory
        that can currently be allocated: at most 4 MB.
        It can become smaller with memory fragmentation.
        The physical allocations in virtual memory can be bigger.
        X = Amount of RAM in use in MB.
        Y = Total amount of RAM available for applications.
        N = The number of free blocks of this size.
        Z = is the size of the largest free block.
    """
    match = RAM_RE.search(text)
    if match:
        return {'use': int(match.group(1)),
                'tot': int(match.group(2)),
                'unit': str(match.group(3)),
                # group 4 is an optional space
                'lfb': {'nblock': int(match.group(5)),
                        'size': int(match.group(6)),
                        'unit': str(match.group(7))}
                }
    else:
        return {}


def MTS(text):
    """ Parse MTS

        MTS fg X% bg Y%
        X = Time spent in foreground tasks.
        Y = Time spent in background tasks.
    """
    match = MTS_RE.search(text)
    if match:
        return {'fg': int(match.group(1)), 'bg': int(match.group(1))}
    else:
        return {}


def VALS(text):
    """ Add all values

        Parse all type of vals:
        - EMC X%@Y
          EMC is the external memory controller,
          through which all sysmem/carve-out/GART memory accesses go.
          X = Percent of EMC memory bandwidth being used, relative to the current running frequency.
          Y = EMC frequency in megahertz.
        - APE Y
          APE is the audio processing engine.
          The APE subsystem consists of ADSP (Cortex®-A9 CPU), mailboxes, AHUB, ADMA, etc.
          Y = APE frequency in megahertz.
        - GR3D X%@Y
          GR3D is the GPU engine.
          X = Percent of the GR3D that is being used, relative to the current running frequency.
          Y = GR3D frequency in megahertz
        - MSENC Y
          Y = MSENC frequency in megahertz.
          MSENC is the video hardware encoding engine.
        - NVENC Y
          Y = NVENC frequency in megahertz.
          NVENC is the video hardware encoding engine.
        - NVDEC Y
          Y = NVDEC frequency in megahertz.
          NVDEC is the video hardware decoding engine.
          It is shown only when hardware decoder/encoder engine is used.
    """
    vals = {}
    for name, val in re.findall(VALS_RE, text):
        # Remove from name "FREQ" name
        name = name.split('_')[0] if "FREQ" in name else name
        # Export value
        vals[name] = val_freq(val)
    return vals


def get_governor(cpus, idx):
    """ Update status governor """
    governor_name = '/sys/devices/system/cpu/cpu' + str(idx - 1) + '/cpufreq/scaling_governor'
    # Add governor CPU if only exist
    if os.path.isfile(governor_name):
        with open(governor_name, 'r') as f:
            cpus['CPU' + str(idx)]['governor'] = f.read()[:-1]


def CPUS(text):
    """ Parse CPU information and extract status

        CPU [X%,Y%, , ]@Z or CPU [X%@Z, Y%@Z,...]
        X and Y are rough approximations based on time spent
        in the system idle process as reported by the Linux kernel in /proc/stat.
        X = Load statistics for each of the CPU cores relative to the
            current running frequency Z, or 'off' in case a core is currently powered down.
        Y = Load statistics for each of the CPU cores relative to the
            current running frequency Z, or 'off' in case a core is currently powered down.
        Z = CPU frequency in megahertz. Goes up or down dynamically depending on the CPU workload.
    """
    match = CPU_RE.search(text)
    cpus = {}
    if match:
        # Extract
        cpus_list = match.group(1).split(',')
        for idx, cpu_str in enumerate(cpus_list):
            # Set name CPU
            name = idx + 1
            cpus[name] = {}
            # status
            if 'off' == cpu_str:
                continue
            # Add data CPU
            val = val_freq(cpu_str)
            cpus[name].update(val)
    return cpus


def TEMPS(text):
    """ Parse all temperatures in tegrastats output

        [temp name]@XC
        [temp name] is one of the names under the nodes
        X = Current temperature
        /sys/devices/virtual/thermal/thermal_zoneX/type.
    """
    return {str(name): float(val) for name, val in re.findall(TEMP_RE, text)}


def WATTS(text):
    """ Parse all milliwatt in tegrastats output

        [VDD_name] X/Y
        X = Current power consumption in milliwatt.
        Y = Average power consumption in milliwatt.
    """
    return {str(name): {'cur': int(cur), 'avg': int(avg)} for name, cur, unit_cur, avg, unit_avg in re.findall(WATT_RE, text)}
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
# Logging
import logging
# Launch command
from .command import Command
from .common import cat, GenericInterface
from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)

COMMAND_TIMEOUT = 4.0
FAN_MANUAL_NAME = 'manual'
FAN_TEMP_CONTROL_NAME = 'temp_control'
FAN_PWM_RE = re.compile(r'^pwm\d+$')
FAN_NVFAN_NAME_RE = re.compile(r'^<FAN (?P<num>\d+)>$')
FAN_NVFAN_OPTIONS_RE = re.compile(r'FAN_(?P<type>\w+) (?P<value>\w+) {$')
FAN_NVFAN_DEFAULT_RE = re.compile(r'FAN_DEFAULT_(?P<type>\w+) (?P<value>\w+)')
# Fan configurations
FAN_PWM_CAP = 255


def ValueToPWM(value, pwm_cap=FAN_PWM_CAP):
    return int(value * pwm_cap / 100)


def PWMtoValue(value, pwm_cap=FAN_PWM_CAP):
    return float(value * 100 / pwm_cap)


def get_all_rpm_system(root_dir):
    pwm_files = {}
    for dir in os.listdir(root_dir):
        full_path = os.path.join(root_dir, dir)
        if os.path.isdir(full_path):
            # Find all pwm in folder
            for file in os.listdir(full_path):
                if 'rpm' == file:
                    name_file = os.path.join(full_path, 'name')
                    name = cat(name_file).strip() if os.path.isfile(name_file) else dir
                    pwm_files[name] = os.path.join(full_path, file)
                    logger.info("RPM {name} found in {root_path}".format(name=name, root_path=full_path))
    return pwm_files


def get_all_cooling_system(root_dir):
    pwm_files = {}
    if not os.path.isdir(root_dir):
        logger.error("Folder {root_dir} doesn't exist".format(root_dir=root_dir))
        return pwm_files
    # Fin all fans
    for dir in os.listdir(root_dir):
        full_path = os.path.join(root_dir, dir)
        if os.path.isdir(full_path):
            fan_device_paths = []
            fan_rpm_path = []
            # Find all pwm in folder
            for file in os.listdir(full_path):
                if FAN_PWM_RE.match(file) or file == 'target_pwm':
                    fan_device_paths += [os.path.join(full_path, file)]
                # Check if there are rpm values
                if file == 'rpm_measured':
                    fan_rpm_path += [os.path.join(full_path, file)]
            # If there are pwm is added in list
            if fan_device_paths:
                name_file = os.path.join(full_path, 'name')
                name = cat(name_file).strip() if os.path.isfile(name_file) else dir
                pwm_files[name] = {'path': full_path, 'pwm': fan_device_paths}
                logger.info("Fan {name}({num}) found in {root_path}".format(name=name, root_path=full_path, num=len(fan_device_paths)))
            if fan_rpm_path:
                pwm_files[name]['rpm'] = fan_rpm_path
                logger.info("RPM {name}({num}) found in {root_path}".format(name=name, root_path=full_path, num=len(fan_device_paths)))
    # Find all rpm systems
    rpm_list = get_all_rpm_system(root_dir)
    for fan, rpm in zip(pwm_files, rpm_list):
        pwm_files[fan]['rpm'] = [rpm_list[rpm]]
    return pwm_files


def get_all_legacy_fan():
    pwm_files = {}
    root_path = ""
    for path in ['/sys/kernel/debug/tegra_fan', '/sys/devices/pwm-fan']:
        if os.path.exists(path):
            root_path = path
            break
    if not root_path:
        return pwm_files
    # Check if this fan is already listed
    if os.path.isdir(os.path.join(root_path, 'hwmon')):
        return pwm_files
    # Otherwise add in list
    name = os.path.basename(root_path)
    # Find all commands
    fan_device_paths = []
    fan_rpm_path = []
    for file in os.listdir(root_path):
        if file == 'target_pwm':
            fan_device_paths += [os.path.join(root_path, file)]
        # Check if there are rpm values
        elif file == 'rpm_measured':
            fan_rpm_path += [os.path.join(root_path, file)]
    # If there are pwm is added in list
    if fan_device_paths:
        pwm_files[name] = {'path': root_path, 'pwm': fan_device_paths}
        logger.info("Found legacy FAN {name}({num}) found in {root_path}".format(name=name, root_path=root_path, num=len(fan_device_paths)))
    if fan_rpm_path:
        pwm_files[name]['rpm'] = fan_rpm_path
        logger.info("Legacy RPM {name}({num}) found in {root_path}".format(name=name, root_path=root_path, num=len(fan_device_paths)))
    return pwm_files


def nvfancontrol_query():
    status = {}
    try:
        nvpmodel_p = Command(['nvfancontrol', '-q'])
        lines = nvpmodel_p(timeout=COMMAND_TIMEOUT)
        for line in lines:
            values = line.split(':')
            fan_name = values[0].lower()
            query = values[1].replace("FAN_", "").lower()
            if fan_name not in status:
                status[fan_name] = {}
            status[fan_name][query] = values[2]
    except (OSError, Command.CommandException):
        pass
    return status


def decode_nvfancontrol():
    nvfan = {}
    current_fan = ''
    if os.path.isfile("/etc/nvfancontrol.conf"):
        with open("/etc/nvfancontrol.conf", 'r') as fp:
            for line in fp:
                match_name = re.search(FAN_NVFAN_NAME_RE, line.strip())
                match_values = re.search(FAN_NVFAN_OPTIONS_RE, line.strip())
                if match_name:
                    parsed_line = match_name.groupdict()
                    current_fan = 'fan{num}'.format(num=parsed_line['num'])
                    nvfan[current_fan] = {}
                elif match_values:
                    parsed_line = match_values.groupdict()
                    type_name = parsed_line['type'].lower()
                    if type_name not in nvfan[current_fan]:
                        nvfan[current_fan][type_name] = []
                    nvfan[current_fan][type_name] += [parsed_line['value']]
    return nvfan


def change_nvfancontrol_default(name, value):
    with open("/etc/nvfancontrol.conf", "r") as f:
        lines = f.readlines()
    with open("/etc/nvfancontrol.conf", "w") as f:
        for line in lines:
            match_defaults = re.search(FAN_NVFAN_DEFAULT_RE, line.strip())
            if match_defaults:
                parsed_line = match_defaults.groupdict()
                if name.upper() == parsed_line['type']:
                    # Override line with new value
                    line = line.replace(parsed_line['value'], value)
            # Print line
            f.write(line)


def nvfancontrol_is_active():
    try:
        output = Command.run_command(['systemctl', 'status', 'nvfancontrol.service'])
        return 'Active: active' in '\n'.join(output)
    except Command.CommandException:
        return False


class Fan(GenericInterface):
    """
    This class enable to control your fan or set of fan.
    Please read the documentation on :py:attr:`~jtop.jtop.fan`

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                jetson.fan.set_profile("tegra_fan", "manual")

    Below all methods available using the :py:attr:`~jtop.jtop.fan` attribute
    """

    def __init__(self):
        super(Fan, self).__init__()
        # list of all profiles in self._init (check services)

    def all_profiles(self, name):
        """
        Return a list of all profiles available for a fan.

        * **Before** Jetpack 5
            * temp_control
            * manual
        * **After** Jetpack 5  jtop map `nvfancontrol <https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/text/SD/PlatformPowerAndPerformance/JetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#fan-profile-control>`_
            * Quiet
            * Cool
            * Manual (Stop nvfancontrol service)

        All fan are always with `Manual` profile option.

        :param name: Name of Fan
        :type name: str
        :raises JtopException: Fan name doesn't exist
        :return: List of all profiles available
        :rtype: list
        """  # noqa
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        return self._init[name]

    def set_profile(self, name, profile):
        """
        Set a new profile for a fan.
        Check which profile is available with :py:func:`~all_profiles`.

        :param name: Name of Fan
        :type name: str
        :param profile: Profile name
        :type profile: str
        :raises JtopException: Fan name doesn't exist or wrong profile
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        if profile not in self.all_profiles(name):
            all_profiles = ' '.join(self.all_profiles(name))
            raise JtopException("Profile \"{profile}\" does not exist for Fan \"{name}\". Available: {all_profiles}".format(
                profile=profile, name=name, all_profiles=all_profiles))
        # Skip if the new profile is the same of the previous
        if profile == self._data[name]['profile']:
            return
        # Set new fan profile
        self._controller.put({'fan': {'command': 'profile', 'name': name, 'profile': profile}})

    def get_profile(self, name):
        """
        Return the current profile is enabled on fan. This value is also readable from :py:attr:`~jtop.jtop.fan`.

        :param name: Name of Fan
        :type name: str
        :raises JtopException: Fan name doesn't exist
        :return: fan profile name
        :rtype: str
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        return self._data[name]['profile']

    def get_profile_default(self, name):
        """
        Get the default profile for this fan. Usually is the first profile listed in :py:func:`~all_profiles`

        :param name: Name of Fan
        :type name: str
        :raises JtopException: Fan name doesn't exist
        :return: profile default name
        :rtype: str
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        return self._init[name][0]

    @property
    def profile(self):
        """
        This property show the current profile selected on first fan on your board.
        This is a simplified version of :py:func:`~set_profile` where *name* is the first fan listed.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Print profile
                    print(jetson.fan.profile)
                    # Set new profile
                    jetson.fan.profile = 'quiet'

        :return: current profile in fan
        :rtype: str
        """
        # Return first fan name and get speed
        if len(self._data) > 0:
            # Extract first name
            name = list(self._data.keys())[0]
            # Get profile
            return self.get_profile(name)
        return None

    @profile.setter
    def profile(self, value):
        if len(self._data) > 0:
            # Extract first name
            name = list(self._data.keys())[0]
            # Set speed for first fan
            self.set_profile(name, value)

    def set_speed(self, name, speed, idx=0):
        """
        Set a new speed for a selected fan.

        :param name: Name of Fan
        :type name: str
        :param speed: New speed value, a number between [0, 100]
        :type speed: float
        :param idx: Index fan, defaults to 0
        :type idx: int, optional
        :raises JtopException: Fan name doesn't exist or wrong index
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        if idx >= len(self._data[name]['speed']) or idx < 0:
            raise JtopException("Fan \"{name}\" have only {len} fans".format(name=name, len=len(self._data[name]['speed'])))
        # Skip if the new profile is the same of the previous
        if speed == self._data[name]['speed'][idx]:
            return
        # Set new fan speed
        self._controller.put({'fan': {'command': 'speed', 'name': name, 'speed': speed, 'idx': idx}})

    def get_speed(self, name, idx=0):
        """
        Return for a selected Fan and index the current speed.
        This value is also readable from :py:attr:`~jtop.jtop.fan`.

        :param name: Name of Fan
        :type name: str
        :param idx: Index fan, defaults to 0
        :type idx: int, optional
        :raises JtopException: Fan name doesn't exist or wrong index
        :return: fan speed a number between [0, 100]
        :rtype: int
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        if idx >= len(self._data[name]['speed']) or idx < 0:
            raise JtopException("Fan \"{name}\" have only {len} fans".format(name=name, len=len(self._data[name]['speed'])))
        return self._data[name]['speed'][idx]

    @property
    def speed(self):
        """
        This property show the current speed between [0, 100] on first fan on your board.
        This is a simplified version of :py:func:`~set_speed` where *name* is the first fan listed.

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read speed
                    print(jetson.fan.speed)
                    # Set new speed
                    jetson.fan.speed = 90

        :return: current fan speed
        :rtype: float
        """
        # Return first fan name and get speed
        if len(self._data) > 0:
            # Extract first name
            name = list(self._data.keys())[0]
            return_value = self.get_speed(name)
            # Return first speed
            return return_value
        return None

    @speed.setter
    def speed(self, value):
        if len(self._data) > 0:
            # Extract first name
            name = list(self._data.keys())[0]
            # Set speed for first fan
            self.set_speed(name, value)

    def get_rpm(self, name, idx=0):
        """
        This method return RPM fan. This output is always available on all Jetson.
        This value is also readable from :py:attr:`~jtop.jtop.fan`.

        :param name: Name of Fan
        :type name: str
        :param idx: Index fan, defaults to 0
        :type idx: int, optional
        :raises JtopException: Fan name doesn't exist, rpm doesn't exist or wrong index
        :return: RPM value
        :rtype: int
        """
        if name not in self._data:
            raise JtopException("Fan \"{name}\" does not exist".format(name=name))
        if 'rpm' not in self._data[name]:
            raise JtopException("Fan \"{name}\" doesn't have RPM".format(name=name))
        if idx >= len(self._data[name]['rpm']) or idx < 0:
            raise JtopException("Fan \"{name}\" have only {len} fans".format(name=name, len=len(self._data[name]['rpm'])))
        return self._data[name]['rpm'][idx]

    @property
    def rpm(self):
        """
        This property show the current fan selected on first fan on your board.

        .. note::

            You can only read RPM, but **not** set a new speed

        You cannot set a new RPM value

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Read RPM
                    print(jetson.fan.rpm)

        :return: rpm first fan
        :rtype: int
        """
        # Return first fan name and get speed
        if len(self._data) > 0:
            # Extract first name
            name = list(self._data.keys())[0]
            # Return first speed
            return self.get_rpm(name)
        return None


def check_config(config):
    if 'fan' not in config:
        return
    fans = config.get('fan')
    for fan in fans:
        if not isinstance(fans[fan], dict):
            logger.warning("Old fan configuration store. Cleaned")
            config.set('fan', {})
        else:
            if 'speed' in fans[fan]:
                if not isinstance(fans[fan]['speed'], list):
                    logger.warning("Old speed configuration store. Cleaned")
                    config.set('fan', {})


class FanService(object):

    def __init__(self, config):
        # Load configuration
        self._config = config
        # Check consistency fan
        check_config(config)
        # Load base hwmon folder
        root_dir = "/sys/class/hwmon"
        if os.getenv('JTOP_TESTING', False):
            root_dir = "/fake_sys/class/hwmon"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=root_dir))
        # Find all fan available
        self._fan_list = get_all_cooling_system(root_dir)
        self._fan_list.update(get_all_legacy_fan())
        # Check if there is nvfan control
        self._nvfancontrol = os.path.isfile('/etc/systemd/system/nvfancontrol.service') or os.path.islink('/etc/systemd/system/nvfancontrol.service')
        # Initialize controller
        if self._nvfancontrol:
            logger.info("Found nvfancontrol.service")
            nv_fan_modes = decode_nvfancontrol()
            # Add all nvfan profiles
            for fan, nvfan in zip(self._fan_list, nv_fan_modes):
                self._fan_list[fan].update(nv_fan_modes[nvfan])
                # Add extra profile for disabled service
                if 'profile' in self._fan_list[fan]:
                    self._fan_list[fan]['profile'] += [FAN_MANUAL_NAME]
        else:
            for name, fan in self._fan_list.items():
                # Initialize profile list
                self._fan_list[name]['profile'] = []
                # Find temp controller
                control = os.path.join(fan['path'], FAN_TEMP_CONTROL_NAME)
                if os.path.isfile(control):
                    # Add control path
                    self._fan_list[name]['control'] = control
                    # Add profiles
                    self._fan_list[name]['profile'] += [FAN_TEMP_CONTROL_NAME]
                    logger.info("Fan temp controller {name} found in {root_path}".format(name=name, root_path=control))
                # Add default profile
                self._fan_list[name]['profile'] += [FAN_MANUAL_NAME]
        if not self._fan_list:
            logger.warning("No fan found")

    def initialization(self):
        # Load configuration
        fan_config = self._config.get('fan', {})
        for name, fan in fan_config.items():
            if 'profile' in fan:
                profile = fan['profile']
                logger.info("Initialization {name}".format(name=name))
                self.set_profile(name, profile)
                if profile == FAN_MANUAL_NAME and 'speed' in fan:
                    speed, index = fan['speed']
                    self.set_speed(name, speed, index)
                    logger.info("Initialization {name} {index} speed {speed}%".format(name=name, index=index, speed=speed))

    def get_configs(self):
        governors = {}
        for fan, data in self._fan_list.items():
            governors[fan] = data['profile']
        return governors

    def get_profile(self, name):
        if name not in self._fan_list:
            logger.error("Fan \"{name}\" does not exist".format(name=name))
            return ""
        profile = FAN_MANUAL_NAME
        if self._nvfancontrol:
            if nvfancontrol_is_active():
                nvfan_query = nvfancontrol_query()
                for fan_list_name, nvfan in zip(self._fan_list, nvfan_query):
                    if fan_list_name == name:
                        return nvfan_query[nvfan]['profile']
        else:
            if 'control' in self._fan_list[name]:
                control_value = int(cat(self._fan_list[name]['control'])) == 1
                return FAN_TEMP_CONTROL_NAME if control_value else FAN_MANUAL_NAME
        return profile

    def set_profile(self, name, profile):
        if name not in self._fan_list:
            logger.error("Fan \"{name}\" does not exist".format(name=name))
            return False
        # Check current status before change
        if profile == self.get_profile(name):
            logger.warning("Fan {name} profile {profile} already active".format(name=name, profile=profile))
            return True
        if self._nvfancontrol:
            is_active = nvfancontrol_is_active()
            # Check first if the fan control is active and after enable the service
            if profile in self._fan_list[name]['profile']:
                if profile == FAN_MANUAL_NAME:
                    if is_active:
                        os.system('systemctl stop nvfancontrol')
                        logger.info("Profile set {profile}".format(profile=profile))
                else:
                    # Check if active and stop
                    if is_active:
                        os.system('systemctl stop nvfancontrol')
                        logger.info("Stop nvfancontrol service")
                    # Update nvfile
                    change_nvfancontrol_default('profile', profile)
                    logger.info("Change /etc/nvfancontrol.conf profile in {profile}".format(profile=profile))
                    # Remove nvfancontrol status file
                    if os.path.isfile("/var/lib/nvfancontrol/status"):
                        os.remove("/var/lib/nvfancontrol/status")
                        logger.info("Removed /var/lib/nvfancontrol/status")
                    # Restart service
                    os.system('systemctl start nvfancontrol')
                    logger.info("Restart nvfancontrol With profile: \"{profile}\"".format(profile=profile))
            else:
                logger.error("Profile {profile} doesn't exist".format(profile=profile))
                return False
        else:
            if profile in self._fan_list[name]['profile']:
                control_value = "0" if profile == FAN_MANUAL_NAME else "1"
                # Write control if exist
                if 'control' in self._fan_list[name]:
                    control = self._fan_list[name]['control']
                    # Set for all pwm the same speed value
                    if os.access(control, os.W_OK):
                        with open(control, 'w') as f:
                            f.write(control_value)
                    logger.info("Profile set {profile}".format(profile=profile))
            else:
                logger.error("Profile {profile} doesn't exist".format(profile=profile))
                return False
        # Update configuration on board
        fan_config = self._config.get('fan', {})
        # Set new profile
        if name not in fan_config:
            fan_config[name] = {}
        fan_config[name]['profile'] = profile
        # Set new jetson_clocks configuration
        self._config.set('fan', fan_config)
        return True

    def set_speed(self, name, speed, index):
        if name not in self._fan_list:
            logger.error("This fan {name} doesn't exist".format(name=name))
            return
        if index >= len(self._fan_list[name]['pwm']):
            logger.error("Wrong index {index} for {name}".format(index=index, name=name))
            return
        # Update configuration on board
        fan_config = self._config.get('fan', {})
        # Set new profile
        if name not in fan_config:
            fan_config[name] = {}
        fan_config[name]['speed'] = (speed, index)
        # Set new jetson_clocks configuration
        self._config.set('fan', fan_config)
        # Convert in PWM
        pwm = ValueToPWM(speed)
        # Set for all pwm the same speed value
        pwm_path = self._fan_list[name]['pwm'][index]
        try:
            if os.access(pwm_path, os.W_OK):
                with open(pwm_path, 'w') as f:
                    f.write(str(pwm))
        except OSError as e:
            logger.error("I cannot set fan speed: {speed} - error {e}".format(speed=speed, e=e))

    def get_status(self):
        fan_status = {}
        # Read all fan status
        for name, data in self._fan_list.items():
            # Read pwm from all fan
            fan_status[name] = {
                'speed': [PWMtoValue(float(cat(pwm))) for pwm in data['pwm']],
            }
            if 'rpm' in data:
                fan_status[name]['rpm'] = [int(cat(rpm)) for rpm in data['rpm']]
        # Check status fan control
        if self._nvfancontrol:
            nvfan_query = {}
            if nvfancontrol_is_active():
                nvfan_query = nvfancontrol_query()
            if nvfan_query:
                for fan, nvfan in zip(fan_status, nvfan_query):
                    fan_status[fan].update(nvfan_query[nvfan])
            else:
                for fan in fan_status:
                    fan_status[fan]['profile'] = FAN_MANUAL_NAME
        else:
            for name, data in self._fan_list.items():
                if 'control' in data:
                    control_value = int(cat(data['control'])) == 1
                    fan_status[name]['profile'] = FAN_TEMP_CONTROL_NAME if control_value else FAN_MANUAL_NAME
                else:
                    fan_status[name]['profile'] = FAN_MANUAL_NAME
        return fan_status
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import curses
# Logging
import logging
# local modules
from ..core.common import get_var
from .lib.colors import NColors
# Create logger
logger = logging.getLogger(__name__)
# Gui refresh rate
COPYRIGHT_RE = re.compile(r""".*__copyright__ = ["'](.*?)['"]""", re.S)
GUI_REFRESH = 1000 // 20

JTOP_MENU_STATUS = 0
JTOP_MENU_EXEC = 1
JTOP_MENU_DESCRIPTION = 2


class JTOPCONFIG:

    def __init__(self, stdscr, main_page):
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        # Initialize variables
        self._current_menu = main_page
        self._previous_menu = main_page
        self._counter_option = 0
        # Initialize keyboard status
        self.key = -1
        self.old_key = -1
        # Initialize mouse
        self.mouse = ()
        # Set curses reference, refresh and jetson controller
        self.stdscr = stdscr
        # In this program, we don't want keystrokes echoed to the console,
        # so we run this to disable that
        curses.noecho()
        # Additionally, we want to make it so that the user does not have to press
        # enter to send keys to our program, so here is how we get keys instantly
        curses.cbreak()
        # Try to hide the cursor
        if hasattr(curses, 'curs_set'):
            try:
                curses.curs_set(0)
            except Exception:
                pass
        # Lastly, keys such as the arrow keys are sent as funny escape sequences to
        # our program. We can make curses give us nicer values (such as curses.KEY_LEFT)
        # so it is easier on us.
        self.stdscr.keypad(True)
        # Enable mouse mask
        _, _ = curses.mousemask(curses.BUTTON1_CLICKED)
        # Refreshing page curses loop
        # https://stackoverflow.com/questions/54409978/python-curses-refreshing-text-with-a-loop
        self.stdscr.nodelay(1)
        # Run loop
        self.loop()

    def draw_menu_page(self, page):
        height, width = self.stdscr.getmaxyx()
        # Draw menu
        menu = []
        if 'menu' in page:
            menu = page['menu']
        # Find center on y axis
        center_y = (height - len(menu)) // 2
        # Find center on X axis
        max_description = max([len(x) for _, _, x in menu]) if len(menu) else 0
        center_x = (width - max_description) // 2
        # Draw menu
        for idx, (status, _, description) in enumerate(menu):
            # Reverse line if counter is the same
            color = curses.A_REVERSE if self._counter_option == idx else curses.A_NORMAL
            # Draw command if not None
            if isinstance(status, str):
                self.stdscr.addstr(center_y + idx, center_x - 9, status, color)
            elif callable(status):
                # Run function and get output
                if 'run_before' in page:
                    cmd_before_out = page['run_before']()
                    status = status(cmd_before_out)
                else:
                    status = status()
                # Print output
                if isinstance(status, bool):
                    message = " OK " if status else "FAIL"
                    color_status = NColors.green() if status else NColors.red()
                    self.stdscr.addstr(center_y + idx, center_x - 6, message, color_status | curses.A_BOLD)
                elif isinstance(status, str):
                    self.stdscr.addstr(center_y + idx, center_x - 6, status)
            # Write description
            self.stdscr.addstr(center_y + idx, center_x, description, color)
        # Draw description
        description_y = 0
        if 'description' in page:
            description = page['description'].split('\n')
            # Find center description
            max_line = max([len(x) for x in description])
            description_center = (max_description - max_line) // 2
            # Draw all lines
            for idx, line in enumerate(description):
                self.stdscr.addstr(center_y - len(description) + idx - 1, center_x + description_center, line)
            description_y = len(description) + 1
        # Draw title
        title = page['title']
        title_center = (max_description - len(title)) // 2
        self.stdscr.addstr(center_y - 2 - description_y, center_x + title_center, title, curses.A_BOLD)
        # Draw buttons
        center_x = (width - 45) // 2
        self.stdscr.addstr(center_y + len(menu) + 1, center_x + 5, "Commands:", curses.A_BOLD)  # Size 9
        if 'menu' in page:
            # Up key
            self.stdscr.addch(center_y + len(menu) + 1, center_x + 30, curses.ACS_UARROW)
            self.stdscr.addstr(center_y + len(menu) + 1, center_x + 32, "Up", curses.A_BOLD)
            # Down key
            self.stdscr.addch(center_y + len(menu) + 2, center_x + 30, curses.ACS_DARROW)
            self.stdscr.addstr(center_y + len(menu) + 2, center_x + 32, "Down", curses.A_BOLD)
            # Enter
            self.stdscr.addstr(center_y + len(menu) + 3, center_x + 30, "ENTER", curses.A_BOLD)
        # Esc button
        message_button = "ESC = Back" if self._current_menu != self._previous_menu else "ESC = exit"
        self.stdscr.addstr(center_y + len(menu) + 1, center_x + 16, "{message}".format(message=message_button), curses.A_BOLD)
        # Quit
        self.stdscr.addstr(center_y + len(menu) + 2, center_x + 16, "q/Q Quit", curses.A_BOLD)
        # Draw copyrights
        self.stdscr.addstr(height - 2, width - 31, "Software parts of jetson-stats", curses.A_NORMAL)
        self.stdscr.addstr(height - 1, width - 28, get_var(COPYRIGHT_RE), curses.A_NORMAL)

    def _print_message(self, title, message):
        return {
            'title': title,
            'description': message
        }

    def loop(self):
        # Here is the loop of our program, we keep clearing and redrawing in this loop
        while not self.events():
            # First, clear the screen
            self.stdscr.erase()
            # Draw menu page
            self.draw_menu_page(self._current_menu)
            # Draw the screen
            self.stdscr.refresh()
            # Set a timeout and read keystroke
            self.stdscr.timeout(GUI_REFRESH)

    def events(self):
        event = self.stdscr.getch()
        # Run keyboard check
        status_keyboard = self.keyboard(event)
        # Clear event mouse
        self.mouse = ()
        # Check event mouse
        if event == curses.KEY_MOUSE:
            try:
                _, mx, my, _, _ = curses.getmouse()
                # Run event menu controller
                self.mouse = (mx, my)
            except curses.error:
                pass
        return status_keyboard

    def keyboard(self, event):
        self.key = event
        if self.old_key != self.key:
            # keyboard check list
            if self.key == ord('q') or self.key == ord('Q'):
                # keyboard check quit button
                return True
            elif self.key == curses.KEY_UP:
                if self._counter_option > 0:
                    self._counter_option -= 1
            elif self.key == curses.KEY_DOWN:
                if self._counter_option + 1 < len(self._current_menu['menu']):
                    self._counter_option += 1
            elif self.key in {curses.KEY_ENTER, 10, 13}:
                cmd = self._current_menu['menu'][self._counter_option][JTOP_MENU_EXEC]
                if isinstance(cmd, dict):
                    self._counter_option = 0
                    self._previous_menu = self._current_menu
                    self._current_menu = cmd
                elif callable(cmd):
                    # Execute command
                    output = cmd()
                    if output:
                        title = self._current_menu['menu'][self._counter_option][JTOP_MENU_DESCRIPTION]
                        data = self._print_message(title, output)
                        self._counter_option = 0
                        self._previous_menu = self._current_menu
                        self._current_menu = data
            elif self.ESC_BUTTON(self.key):
                if self._current_menu != self._previous_menu:
                    self._current_menu = self._previous_menu
                    self._counter_option = 0
                else:
                    return True
            # Store old value key
            self.old_key = self.key
        return False

    def ESC_BUTTON(self, key):
        """
            Check there is another character prevent combination ALT + <OTHER CHR>
            https://stackoverflow.com/questions/5977395/ncurses-and-esc-alt-keys
        """
        if key == 27:
            n = self.stdscr.getch()
            if n == -1:
                return True
        return False
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
# Logging
import logging
from .common import cat, GenericInterface
from .exceptions import JtopException
from .command import Command
# Create logger
logger = logging.getLogger(__name__)
# default ipgu path for Jetson devices
DEFAULT_IGPU_PATH = "/sys/class/devfreq/"


def check_nvidia_smi():
    cmd = Command(['nvidia-smi'])
    try:
        cmd()
        return True
    except (OSError, Command.CommandException):
        pass
    return False


def igpu_read_freq(path):
    # Read status online
    gpu = {}
    # Check if access to this file
    if os.access(path + "/governor", os.R_OK):
        with open(path + "/governor", 'r') as f:
            # Write current engine
            gpu['governor'] = f.read().strip()
    # Check if access to this file
    if os.access(path + "/cur_freq", os.R_OK):
        with open(path + "/cur_freq", 'r') as f:
            # Write current engine
            gpu['cur'] = int(f.read()) // 1000
    # Decode clock rate
    if os.access(path + "/max_freq", os.R_OK):
        with open(path + "/max_freq", 'r') as f:
            # Write status engine
            gpu['max'] = int(f.read()) // 1000
    if os.access(path + "/min_freq", os.R_OK):
        with open(path + "/min_freq", 'r') as f:
            # Write status engine
            gpu['min'] = int(f.read()) // 1000
    # Read GPC status
    for idx in range(2):
        # Read power control status
        path_gpc = "/sys/kernel/debug/bpmp/debug/clk/nafll_gpc{number}/pto_counter".format(number=idx)
        if os.access(path_gpc, os.R_OK):
            with open(path_gpc, 'r') as f:
                # List all frequencies
                if 'GPC' not in gpu:
                    gpu['GPC'] = []
                gpu['GPC'] += [int(f.read()) // 1000]
    return gpu


def igpu_read_status(path):
    gpu = {}
    # GPU status
    if os.access(path + "/railgate_enable", os.R_OK):
        with open(path + "/railgate_enable", 'r') as f:
            # Read status railgate
            gpu['railgate'] = int(f.read()) == 1
    # Mask status (Useful for nvpmodel)
    if os.access(path + "/tpc_pg_mask", os.R_OK):
        with open(path + "/tpc_pg_mask", 'r') as f:
            # Read status TPG PG mask
            gpu['tpc_pg_mask'] = int(f.read()) == 1
    # Status 3D scaling
    if os.access(path + "/enable_3d_scaling", os.R_OK):
        with open(path + "/enable_3d_scaling", 'r') as f:
            # Read status 3D scaling
            gpu['3d_scaling'] = int(f.read()) == 1
    # Current load GPU
    if os.access(path + "/load", os.R_OK):
        with open(path + "/load", 'r') as f:
            # Read current GPU load
            gpu['load'] = float(f.read()) / 10.0
    return gpu


def get_raw_igpu_devices():
    igpu_path = DEFAULT_IGPU_PATH
    raw_output = {}
    for item in os.listdir(igpu_path):
        item_path = os.path.join(igpu_path, item)
        if os.path.isfile(item_path) or os.path.islink(item_path):
            # Check name device
            name_path = "{item}/device/of_node/name".format(item=item_path)
            if os.path.isfile(name_path):
                # Decode name
                name = cat(name_path)
                # path and file
                raw_output[name_path] = "{}".format(name)
    return raw_output


def find_igpu(igpu_path):
    # Check if exist a integrated gpu
    # if not os.path.exists("/dev/nvhost-gpu") and not os.path.exists("/dev/nvhost-power-gpu"):
    #     return []
    igpu = {}
    if not os.path.isdir(igpu_path):
        logger.error("Folder {root_dir} doesn't exist".format(root_dir=igpu_path))
        return igpu
    for item in os.listdir(igpu_path):
        item_path = os.path.join(igpu_path, item)
        if os.path.isfile(item_path) or os.path.islink(item_path):
            # Check name device
            name_path = "{item}/device/of_node/name".format(item=item_path)
            if os.path.isfile(name_path):
                # Decode name
                name = cat(name_path)
                # Check if gpu
                if name in ['gv11b', 'gp10b', 'ga10b', 'gpu']:
                    # Extract real path GPU device
                    path = os.path.realpath(os.path.join(item_path, "device"))
                    frq_path = os.path.realpath(item_path)
                    igpu[name] = {'type': 'integrated', 'path': path, 'frq_path': frq_path}
                    logger.info("GPU \"{name}\" status in {path}".format(name=name, path=path))
                    logger.info("GPU \"{name}\" frq in {path}".format(name=name, path=frq_path))
                    # Check if railgate exist
                    path_railgate = os.path.join(path, "railgate_enable")
                    if os.path.isfile(path_railgate):
                        igpu[name]['railgate'] = path_railgate
                    # Check if 3d scaling exist
                    path_3d_scaling = os.path.join(path, "enable_3d_scaling")
                    if os.path.isfile(path_3d_scaling):
                        igpu[name]['3d_scaling'] = path_3d_scaling
                else:
                    logger.debug("Skipped {name}".format(name=name))
    return igpu


def find_dgpu():
    # Check if there are discrete gpu
    # if not os.path.exists("/dev/nvidiactl") and not os.path.isdir("/dev/nvgpu-pci"):
    #     return []
    # https://enterprise-support.nvidia.com/s/article/Useful-nvidia-smi-Queries-2
    dgpu = {}
    if check_nvidia_smi():
        logger.info("NVIDIA SMI exist!")
    if dgpu:
        logger.info("Discrete GPU found")
    return dgpu


class GPU(GenericInterface):
    """
    This class get the output from your GPU, this class is readable like a dictionary,
    please read the documentation on :py:attr:`~jtop.jtop.gpu` but is also usable to enable, disable 3d scaling on your device.

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                jetson.gpu.set_scaling_3D("gpu", True)

    Below all methods available using the :py:attr:`~jtop.jtop.gpu` attribute
    """

    def __init__(self):
        super(GPU, self).__init__()

    def set_scaling_3D(self, name, value):
        """
        Enable disable GPU 3D scaling. this method send a command like below on your Jetson.

        Set 3D scaling on your board, like the command below. To know the GPU name use :py:attr:`~jtop.jtop.gpu`

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    jetson.gpu.set_scaling_3D("ga10b", True)

        is equivalent to:

        .. code-block:: bash
            :class: no-copybutton

            echo 1 > /sys/devices/17000000.ga10b/enable_3d_scaling

        :param name: GPU name
        :type name: str
        :param value: Enable/Disable 3D scaling
        :type value: bool
        :raises JtopException: if GPU doesn't exist
        """
        if name not in self._data:
            raise JtopException("GPU \"{name}\" does not exist".format(name=name))
        # Set new 3D scaling
        self._controller.put({'gpu': {'command': '3d_scaling', 'name': name, 'value': value}})

    def get_scaling_3D(self, name):
        """
        Return status of 3D scaling, this output is also readable from :py:attr:`~jtop.jtop.gpu` attribute

        :param name: GPU name
        :type name: str
        :raises JtopException: if GPU doesn't exist
        :return: status 3D scaling
        :rtype: bool
        """
        if name not in self._data:
            raise JtopException("GPU \"{name}\" does not exist".format(name=name))
        return self._data[name]['status']['3d_scaling']

    @property
    def scaling_3D(self):
        """
        Return status of 3D scaling, this output is also readable from :py:attr:`~jtop.jtop.gpu` attribute

        .. code-block:: python

            with jtop() as jetson:
                if jetson.ok():
                    # Set new 3D scaling
                    jetson.gpu.set_scaling_3D = True
                    # same of
                    jetson.gpu.set_scaling_3D("ga10b", True)

        :raises JtopException: if there are no integrated GPU
        :return: status 3D scaling
        :rtype: bool
        """
        # Get first integrated gpu
        name = self._get_first_integrated_gpu()
        if not name:
            raise JtopException("no Integrated GPU available")
        return self.get_scaling_3D(name)

    @scaling_3D.setter
    def scaling_3D(self, value):
        # Get first integrated gpu
        name = self._get_first_integrated_gpu()
        if not name:
            raise JtopException("no Integrated GPU available")
        self.set_scaling_3D(name, value)

    def set_railgate(self, name, value):
        if name not in self._data:
            raise JtopException("GPU \"{name}\" does not exist".format(name=name))
        # Set new 3D scaling
        self._controller.put({'gpu': {'command': 'railgate', 'name': name, 'value': value}})

    def get_railgate(self, name):
        if name not in self._data:
            raise JtopException("GPU \"{name}\" does not exist".format(name=name))
        return self._data[name]['status']['railgate']

    def _get_first_integrated_gpu(self):
        for name in self._data:
            if self._data[name]['type'] == 'integrated':
                return name
        return ''


class GPUService(object):

    def __init__(self):
        # Detect integrated GPU
        igpu_path = DEFAULT_IGPU_PATH
        if os.getenv('JTOP_TESTING', False):
            igpu_path = "/fake_sys/class/devfreq/"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=igpu_path))
        self._gpu_list = find_igpu(igpu_path)
        # Find discrete GPU
        self._gpu_list.update(find_dgpu())
        # Check status
        if not self._gpu_list:
            logger.warning("No NVIDIA GPU available")

    def set_scaling_3D(self, name, value):
        if name not in self._gpu_list:
            logger.error("GPU \"{name}\" does not exist".format(name=name))
            return False
        if '3d_scaling' not in self._gpu_list[name]:
            logger.error("GPU \"{name}\" does not have 3D scaling".format(name=name))
            return False
        path_3d_scaling = self._gpu_list[name]['3d_scaling']
        string_value = "1" if value else "0"
        # Write new status 3D scaling
        try:
            if os.access(path_3d_scaling, os.W_OK):
                with open(path_3d_scaling, 'w') as f:
                    f.write(string_value)
            logger.info("GPU \"{name}\" set 3D scaling to {value}".format(name=name, value=value))
        except OSError as e:
            logger.error("I cannot set 3D scaling {}".format(e))

    def set_railgate(self, name, value):
        if name not in self._gpu_list:
            logger.error("GPU \"{name}\" does not exist".format(name=name))
            return False
        if 'railgate' not in self._gpu_list[name]:
            logger.error("GPU \"{name}\" does not have railgate".format(name=name))
            return False
        path_railgate = self._gpu_list[name]['railgate']
        string_value = "1" if value else "0"
        # Write new status railgate
        try:
            if os.access(path_railgate, os.W_OK):
                with open(path_railgate, 'w') as f:
                    f.write(string_value)
            logger.info("GPU \"{name}\" set railgate to {value}".format(name=name, value=value))
        except OSError as e:
            logger.error("I cannot set Railgate {}".format(e))

    def get_status(self):
        gpu_list = {}
        # Read iGPU frequency
        for name, data in self._gpu_list.items():
            # Initialize GPU status
            gpu = {'type': data['type']}
            # Detect frequency and load
            if gpu['type'] == 'integrated':
                # Read status GPU
                gpu['status'] = igpu_read_status(data['path'])
                # Read frequency
                gpu['freq'] = igpu_read_freq(data['frq_path'])
                # Read power control status
                if os.access(data['path'] + "/power/control", os.R_OK):
                    with open(data['path'] + "/power/control", 'r') as f:
                        gpu['power_control'] = f.read().strip()
            elif gpu['type'] == 'discrete':
                logger.info("TODO discrete GPU")
            # Load all status in GPU
            gpu_list[name] = gpu
        return gpu_list
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Logging
import logging
import sys
# Launch command
import subprocess as sp
# Threading
from threading import Thread, Event
# Tegrastats parser
from .tegra_parse import DATE, VALS, MTS, RAM, SWAP, IRAM, CPUS, TEMPS, WATTS
from .common import locate_commands
# Create logger for tegrastats
logger = logging.getLogger(__name__)


class Tegrastats:
    """
        - Subprocess read:
        https://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python/4896288#4896288
        - Property
        https://www.programiz.com/python-programming/property
    """

    def __init__(self, callback, tegrastats_path):
        self._running = Event()
        # Error message from thread
        self._error = None
        # Start process tegrastats
        self.path = locate_commands("tegrastats", tegrastats_path)
        # Define Tegrastats process
        self._thread = None
        # Initialize callback
        self.callback = callback

    def _decode(self, text):
        # Remove date from tegrastats string
        text = DATE(text)
        # Find and parse all single values
        stats = VALS(text)
        # Parse if exist MTS
        mts = MTS(text)
        if mts:
            stats['MTS'] = mts
        # Parse RAM
        stats['RAM'] = RAM(text)
        # If exists parse SWAP
        swap = SWAP(text)
        if swap:
            stats['SWAP'] = swap
        # If exists parse IRAM
        iram = IRAM(text)
        if iram:
            stats['IRAM'] = iram
        # Parse CPU status
        stats['CPU'] = CPUS(text)
        # Parse temperatures
        stats['TEMP'] = TEMPS(text)
        # Parse Watts
        stats['WATT'] = WATTS(text)
        return stats

    def _read_tegrastats(self, interval, running):
        pts = sp.Popen([self.path, '--interval', str(interval)], stdout=sp.PIPE)
        try:
            # Reading loop
            while running.is_set():
                if pts.poll() is not None:
                    continue
                out = pts.stdout
                if out is not None:
                    # Read line process output
                    line = out.readline()
                    # Decode line in UTF-8
                    tegrastats_data = line.decode("utf-8")
                    # Decode and store
                    stats = self._decode(tegrastats_data)
                    # Launch callback
                    self.callback(stats)
        except AttributeError:
            pass
        except IOError:
            pass
        except Exception:
            # Write error message
            self._error = sys.exc_info()
        finally:
            # Kill process
            try:
                pts.kill()
            except OSError:
                pass

    def open(self, interval=0.5):
        if self._thread is not None:
            return False
        # Set timeout
        interval = int(interval * 1000)
        # Check if thread or process exist
        self._running.set()
        # Start thread Service client
        self._thread = Thread(target=self._read_tegrastats, args=(interval, self._running, ))
        self._thread.start()
        return True

    def close(self, timeout=None):
        # Catch exception if exist
        if self._error:
            # Extract exception and raise
            ex_type, ex_value, tb_str = self._error
            ex_value.__traceback__ = tb_str
            raise ex_value
        # Check if thread and process are already empty
        self._running.clear()
        if self._thread is not None:
            self._thread.join(timeout)
            self._thread = None
        return True
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Reference:
# 1. https://packaging.python.org
# 2. https://julien.danjou.info/starting-your-first-python-project/
# 3. https://medium.com/@trstringer/the-easy-and-nice-way-to-do-cli-apps-in-python-5d9964dc950d
# 4. https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/
# 5. https://python-packaging.readthedocs.io
# 6. https://github.com/pypa/sampleproject
# 7. https://pypi.org/classifiers/
# 8. https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from jtop.service import status_service, remove_service_pipe, uninstall_service, set_service_permission, unset_service_permission, install_service
from jtop.core.jetson_variables import uninstall_variables, install_variables
from jtop.terminal_colors import bcolors
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open
# Launch command
import os
import sys
import re
import logging
import shutil
logging.basicConfig(stream=sys.stderr, level=logging.INFO)
log = logging.getLogger()


here = os.path.abspath(os.path.dirname(__file__))
project_homepage = "https://github.com/rbonghi/jetson_stats"
documentation_homepage = "https://rnext.it/jetson_stats"

# Load requirements
with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Load version package
with open(os.path.join(here, "jtop", "__init__.py")) as fp:
    VERSION = (
        re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S).match(fp.read()).group(1)
    )
# Store version package
version = VERSION


def is_virtualenv():
    # https://stackoverflow.com/questions/1871549/determine-if-python-is-running-inside-virtualenv
    if os.path.exists(os.path.join(sys.prefix, 'conda-meta')):
        # Conda virtual environments
        return True
    if hasattr(sys, 'real_prefix'):
        return True
    if hasattr(sys, 'base_prefix'):
        return sys.prefix != sys.base_prefix
    return False


def is_docker():
    # https://gist.github.com/anantkamath/623ce7f5432680749e087cf8cfba9b69
    # https://stackoverflow.com/questions/68816329/how-to-get-docker-container-id-from-within-the-container-with-cgroup-v2
    def check_mountinfo():
        with open('/proc/self/mountinfo', 'r') as file:
            line = file.readline().strip()
            while line:
                if '/docker/containers/' in line or '/docker/volumes/buildx_buildkit_builder' in line:
                    return True
                line = file.readline().strip()
        return False
    # Check on cgroup
    with open('/proc/self/cgroup', 'r') as procfile:
        for line in procfile:
            # if is the new cgroup v2 check on mountinfo
            if line.startswith("0::/"):
                return check_mountinfo()
            fields = line.strip().split('/')
            if 'docker' in fields or 'buildkit' in fields:
                return True
    return False


def is_superuser():
    return os.getuid() == 0


def remove_data(file_name):
    # Remove old pipes if exists
    if os.path.isfile(file_name):
        print("Remove {file} file".format(file=file_name))
        os.remove(file_name)
    elif os.path.isdir(file_name):
        print("Remove {file} folder".format(file=file_name))
        shutil.rmtree(file_name)


def remove_deprecated_data():
    """
    This function uninstall the service
    """
    # If exist, remove old services names if they exists
    uninstall_service('jetson_performance.service')
    uninstall_service('jetson_stats_boot.service')
    uninstall_service('jetson_stats.service')
    # Remove old variable definitions
    uninstall_variables('jetson_env.sh')
    # Remove old permission and group
    unset_service_permission('jetson_stats')
    # Remove old script if they exists
    remove_data("/usr/local/bin/jetson-docker")
    remove_data("/usr/local/bin/jetson-release")
    # Remove old folders
    remove_data("/usr/local/jetson_stats")
    remove_data("/opt/jetson_stats")
    remove_data("/etc/jetson-swap")
    remove_data("/etc/jetson_easy")


def pypi_installer(installer, obj, copy):
    print("Install status:")
    print(" - [{status}] super_user".format(status="X" if is_superuser() else " "))
    print(" - [{status}] virtualenv".format(status="X" if is_virtualenv() else " "))
    print(" - [{status}] docker".format(status="X" if is_docker() else " "))
    # Run the uninstaller before to copy all scripts
    if not is_virtualenv() and not is_docker():
        if is_superuser():
            # Remove all deprecated data
            # - This function should do nothing
            remove_deprecated_data()
            # remove service jtop.service
            uninstall_service()
            # Remove service path
            remove_service_pipe()
            # Uninstall variables
            uninstall_variables()
        else:
            print("----------------------------------------")
            print("Install on your host using superuser permission, like:")
            print(bcolors.bold("sudo pip3 install -U jetson-stats"))
            sys.exit(1)
    elif is_docker():
        print("Skip uninstall in docker")
    else:
        if is_superuser():
            print("Skip uninstall on virtual environment")
        elif not status_service():
            print("----------------------------------------")
            print("Please, before install in your virtual environment, install jetson-stats on your host with superuser permission, like:")
            print(bcolors.bold("sudo pip3 install -U jetson-stats"))
            sys.exit(1)
    # Run the default installation script
    installer.run(obj)
    # Run the restart all services before to close the installer
    if not is_virtualenv() and not is_docker() and is_superuser():
        folder, _ = os.path.split(os.path.realpath(__file__))  # This folder
        # Install variables
        install_variables(folder, copy=copy)
        # Set service permissions
        set_service_permission()
        # Install service (linking only for develop)
        install_service(folder, copy=copy)
    else:
        print("Skip install service")


class JTOPInstallCommand(install):
    """Installation mode."""

    def run(self):
        # Run the uninstaller before to copy all scripts
        pypi_installer(install, self, True)


class JTOPDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        # Run the uninstaller before to copy all scripts
        # Install services (linking)
        pypi_installer(develop, self, False)


# Configuration setup module
setup(
    name="jetson-stats",
    version=version,
    author="Raffaello Bonghi",
    author_email="raffaello@rnext.it",
    description="Interactive system-monitor and process viewer for all NVIDIA Jetson [Orin, Xavier, Nano, TX] series",
    license='AGPL-3.0',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=documentation_homepage,
    project_urls={
        'Documentation': documentation_homepage,
        'Funding': 'https://github.com/sponsors/rbonghi',
        'Say Thanks!': 'https://discord.gg/BFbuJNhYzS',
        'Source': project_homepage,
        'Tracker': (project_homepage + "/issues"),
        'Examples': (project_homepage + "/tree/master/examples"),
    },
    packages=find_packages(exclude=['examples', 'scripts', 'tests', 'jtop.tests', 'jtop.tests_gui']),  # Required
    # Define research keywords
    keywords=("jetson_stats jtop python system-monitor docker \
               nvidia Jetson Orin AGXOrin Xavier AGXXavier XavierNX Nano TX1 TX2 process viewer"
              ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        # Audience and topics
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Debuggers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: System :: Hardware",
        "Topic :: System :: Logging",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Operating System",
        "Topic :: System :: Operating System Kernels",
        "Topic :: System :: Shells",
        "Topic :: System :: Systems Administration",
        "Topic :: Terminals",
        # License
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        # Programming and Operative system
        "Programming Language :: Unix Shell",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX :: Linux"],
    # Requisites
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=2.7',
    platforms=["linux", "linux2", "darwin"],
    install_requires=requirements,
    # Zip safe configuration
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#setting-the-zip-safe-flag
    zip_safe=False,
    # Add jetson_variables in /opt/jetson_stats
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    data_files=[('jetson_stats', ['services/jtop.service', 'scripts/jtop_env.sh'])],
    # Install extra scripts
    cmdclass={'develop': JTOPDevelopCommand,
              'install': JTOPInstallCommand},
    # The following provide a command called `jtop`
    entry_points={'console_scripts': [
        'jtop=jtop.__main__:main',
        'jetson_release = jtop.jetson_release:main',
        'jetson_config = jtop.jetson_config:main',
        'jetson_swap = jtop.jetson_swap:main',
    ]},
)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
import time
from copy import deepcopy


class SmallButton(object):
    def __init__(self, stdscr, on_click, label='', toggle=False, trigger_key=None, info={}):
        self.stdscr = stdscr
        self.on_click = on_click
        self.toggle = toggle
        self.selected = False
        self.trigger_key = trigger_key
        self.info = info
        self.label = label
        self.highlight_time = None

    def update(self, y, x, label='', key=None, mouse=None, color=None):
        label = label if label else self.label
        if mouse and mouse[1] == y and x <= mouse[0] <= x + len(label) + 1:
            self.selected = not self.selected if self.toggle else True
            self.on_click(self.info, selected=self.selected)
            self.highlight_time = time.time()
        elif self.trigger_key:
            if key == ord(self.trigger_key):
                self.selected = not self.selected if self.toggle else True
                self.on_click(self.info, selected=self.selected)
                self.highlight_time = time.time()

        color_normal = color if color else curses.A_NORMAL

        if self.trigger_key:
            # self.stdscr.addstr(y, x, '[{trigger_key}]{label}'.format(trigger_key=self.trigger_key,
            #                   label=" " + label if label else ""), curses.A_REVERSE if self.selected else color_normal)
            self.stdscr.addstr(y, x, '[', curses.A_REVERSE if self.selected else curses.A_NORMAL)
            self.stdscr.addstr('{trigger_key}'.format(trigger_key=self.trigger_key), curses.A_REVERSE if self.selected else color_normal)
            if label:
                self.stdscr.addstr('|', curses.A_REVERSE if self.selected else curses.A_NORMAL)
            self.stdscr.addstr('{label}'.format(label=" " + label if label else ""), curses.A_REVERSE if self.selected else color_normal)
            self.stdscr.addstr(']', curses.A_REVERSE if self.selected else curses.A_NORMAL)
        else:
            # self.stdscr.addstr(y, x, '[{label}]'.format(label=label), curses.A_REVERSE if self.selected else color_normal)
            self.stdscr.addstr(y, x, '[', curses.A_REVERSE if self.selected else curses.A_NORMAL)
            self.stdscr.addstr('{label}'.format(label=label), curses.A_REVERSE if self.selected else color_normal)
            self.stdscr.addstr(']', curses.A_REVERSE if self.selected else curses.A_NORMAL)
        if self.highlight_time and not self.toggle and time.time() - self.highlight_time > 0.1:
            self.selected = False
            self.highlight_time = None

    def is_selected(self):
        return self.selected

    def set_selected(self, value):
        self.selected = value

    def get_label(self):
        return self.label


class HideButton(SmallButton):

    def __init__(self, stdscr, key, text):
        hide_message = "XX CLICK TO READ XXX"
        super(HideButton, self).__init__(stdscr, self.action_on_click, trigger_key=key, label=hide_message)
        self._text = text
        self._pressed = False

    def action_on_click(self, info, selected):
        self.label = self._text
        self._pressed = True

    def update(self, y, x, label='', key=None, mouse=None, color=None):
        if mouse and mouse[1] == y and x <= mouse[0] <= x + len(self.label) + 1:
            self.selected = not self.selected if self.toggle else True
            self.on_click(self.info, selected=self.selected)
            self.highlight_time = time.time()
        if key == ord(self.trigger_key):
            self.selected = not self.selected if self.toggle else True
            self.on_click(self.info, selected=self.selected)
            self.highlight_time = time.time()

        try:
            if not self._pressed:
                self.stdscr.addstr(y, x, '[{key}|{label}]'.format(key=self.trigger_key, label=self.label),
                                   curses.A_REVERSE if self.selected else curses.A_NORMAL)
            else:
                self.stdscr.addstr(y, x, '{label}'.format(label=self.label), curses.A_NORMAL)
        except curses.error:
            pass

        if self.highlight_time and not self.toggle and time.time() - self.highlight_time > 0.1:
            self.selected = False
            self.highlight_time = None


class ButtonList:
    def __init__(self, stdscr, on_click, buttons=[], info={}, linear=False):
        self.buttons = []
        self.on_click = on_click
        self.stdscr = stdscr
        self.info = info
        self._selected = ''
        self.linear = linear
        # Load all buttons
        for button in buttons:
            self.add_button(button)

    def add_button(self, label):
        info_button = deepcopy(self.info)
        info_button['label'] = label
        button = SmallButton(self.stdscr, self.on_click, toggle=True, label=label, info=info_button)
        self.buttons.append(button)

    def update(self, y, x, key, mouse, selected_button, colors=[]):
        # Update selected button
        self._selected = selected_button
        # read color status
        if not colors:
            colors = [None] * len(self.buttons)
        # Show all buttons
        counter_space = 0
        for i, (button, color) in enumerate(zip(self.buttons, colors)):
            name = button.get_label()
            button.set_selected(self._selected == name)
            if self.linear:
                button.update(y, x + counter_space, key=key, mouse=mouse, color=color)
                counter_space += len(name) + 3
            else:
                button.update(y + i, x, key=key, mouse=mouse, color=color)
# EOF
